@@ -33,10 +33,10 @@
 
 #include <string.h>
 
+#undef DEBUG_TOOLTIP
+
 
-#define GTK_TOOLTIP(obj)                 (G_TYPE_CHECK_INSTANCE_CAST ((obj), GTK_TYPE_TOOLTIP, GtkTooltip))
 #define GTK_TOOLTIP_CLASS(klass)         (G_TYPE_CHECK_CLASS_CAST ((klass), GTK_TYPE_TOOLTIP, GtkTooltipClass))
-#define GTK_IS_TOOLTIP(obj)              (G_TYPE_CHECK_INSTANCE_TYPE ((obj), GTK_TYPE_TOOLTIP))
 #define GTK_IS_TOOLTIP_CLASS(klass)      (G_TYPE_CHECK_CLASS_TYPE ((klass), GTK_TYPE_TOOLTIP))
 #define GTK_TOOLTIP_GET_CLASS(obj)       (G_TYPE_INSTANCE_GET_CLASS ((obj), GTK_TYPE_TOOLTIP, GtkTooltipClass))
 
@@ -66,8 +66,11 @@ struct _GtkTooltip
   guint timeout_id;
   guint browse_mode_timeout_id;
 
+  GdkRectangle tip_area;
+
   guint browse_mode_enabled : 1;
   guint keyboard_mode_enabled : 1;
+  guint tip_area_set : 1;
 };
 
 struct _GtkTooltipClass
@@ -82,6 +85,7 @@ static void       gtk_tooltip_class_init           (GtkTooltipClass *klass);
 static void       gtk_tooltip_init                 (GtkTooltip      *tooltip);
 static void       gtk_tooltip_finalize             (GObject         *object);
 
+static void       gtk_tooltip_window_style_set     (GtkTooltip      *tooltip);
 static gboolean   gtk_tooltip_paint_window         (GtkTooltip      *tooltip);
 static void       gtk_tooltip_window_hide          (GtkWidget       *widget,
 						    gpointer         user_data);
@@ -137,6 +141,8 @@ gtk_tooltip_init (GtkTooltip *tooltip)
   gtk_container_add (GTK_CONTAINER (tooltip->window), tooltip->alignment);
   gtk_widget_show (tooltip->alignment);
 
+  g_signal_connect_swapped (tooltip->window, "style_set",
+			    G_CALLBACK (gtk_tooltip_window_style_set), tooltip);
   g_signal_connect_swapped (tooltip->window, "expose_event",
 			    G_CALLBACK (gtk_tooltip_paint_window), tooltip);
 
@@ -149,6 +155,7 @@ gtk_tooltip_init (GtkTooltip *tooltip)
 		      FALSE, FALSE, 0);
 
   tooltip->label = gtk_label_new ("");
+  gtk_label_set_line_wrap (GTK_LABEL (tooltip->label), TRUE);
   gtk_box_pack_start (GTK_BOX (tooltip->box), tooltip->label,
 		      FALSE, FALSE, 0);
 
@@ -313,9 +320,14 @@ gtk_tooltip_set_custom (GtkTooltip *tooltip,
 
   if (tooltip->custom_widget)
     {
-      gtk_container_remove (GTK_CONTAINER (tooltip->box),
-			    tooltip->custom_widget);
-      g_object_unref (tooltip->custom_widget);
+      GtkWidget *custom = tooltip->custom_widget;
+      /* Note: We must reset tooltip->custom_widget first, 
+       * since gtk_container_remove() will recurse into 
+       * gtk_tooltip_set_custom()
+       */
+      tooltip->custom_widget = NULL;
+      gtk_container_remove (GTK_CONTAINER (tooltip->box), custom);
+      g_object_unref (custom);
     }
 
   if (custom_widget)
@@ -325,13 +337,42 @@ gtk_tooltip_set_custom (GtkTooltip *tooltip,
       gtk_container_add (GTK_CONTAINER (tooltip->box), custom_widget);
       gtk_widget_show (custom_widget);
     }
+}
+
+/**
+ * gtk_tooltip_set_tip_area:
+ * @tooltip: a #GtkTooltip
+ * @rect: a #GdkRectangle
+ *
+ * Sets the area of the widget, where the contents of this tooltip apply,
+ * to be @rect (in widget coordinates).  This is especially useful for
+ * properly setting tooltips on #GtkTreeView rows and cells, #GtkIconViews,
+ * etc.
+ *
+ * For setting tooltips on #GtkTreeView, please refer to the convenience
+ * functions for this: gtk_tree_view_set_tooltip_row() and
+ * gtk_tree_view_set_tooltip_cell().
+ *
+ * Since: 2.12
+ */
+void
+gtk_tooltip_set_tip_area (GtkTooltip   *tooltip,
+			  GdkRectangle *rect)
+{
+  g_return_if_fail (GTK_IS_TOOLTIP (tooltip));
+
+  if (!rect)
+    tooltip->tip_area_set = FALSE;
   else
-    tooltip->custom_widget = NULL;
+    {
+      tooltip->tip_area_set = TRUE;
+      tooltip->tip_area = *rect;
+    }
 }
 
 /**
  * gtk_tooltip_trigger_tooltip_query:
- * @display: a #GtkDisplay
+ * @display: a #GdkDisplay
  *
  * Triggers a new tooltip query on @display, in order to update the current
  * visible tooltip, or to show/hide the current tooltip.  This function is
@@ -369,6 +410,19 @@ gtk_tooltip_reset (GtkTooltip *tooltip)
   gtk_tooltip_set_markup (tooltip, NULL);
   gtk_tooltip_set_icon (tooltip, NULL);
   gtk_tooltip_set_custom (tooltip, NULL);
+  gtk_tooltip_set_tip_area (tooltip, NULL);
+}
+
+static void
+gtk_tooltip_window_style_set (GtkTooltip *tooltip)
+{
+  gtk_alignment_set_padding (GTK_ALIGNMENT (tooltip->alignment),
+			     tooltip->window->style->ythickness,
+			     tooltip->window->style->ythickness,
+			     tooltip->window->style->xthickness,
+			     tooltip->window->style->xthickness);
+
+  gtk_widget_queue_draw (tooltip->window);
 }
 
 static gboolean
@@ -394,8 +448,7 @@ gtk_tooltip_window_hide (GtkWidget *widget,
 {
   GtkTooltip *tooltip = GTK_TOOLTIP (user_data);
 
-  if (tooltip->custom_widget)
-    gtk_tooltip_set_custom (tooltip, NULL);
+  gtk_tooltip_set_custom (tooltip, NULL);
 }
 
 /* event handling, etc */
@@ -416,11 +469,31 @@ child_location_foreach (GtkWidget *child,
   gint x, y;
   struct ChildLocation *child_loc = data;
 
+  /* Ignore invisible widgets */
+  if (!GTK_WIDGET_DRAWABLE (child))
+    return;
+
+  /* (child_loc->x, child_loc->y) are relative to
+   * child_loc->container's allocation.
+   */
+
   if (!child_loc->child &&
       gtk_widget_translate_coordinates (child_loc->container, child,
 					child_loc->x, child_loc->y,
 					&x, &y))
     {
+#ifdef DEBUG_TOOLTIP
+      g_print ("candidate: %s  alloc=[(%d,%d)  %dx%d]     (%d, %d)->(%d, %d)\n",
+	       gtk_widget_get_name (child),
+	       child->allocation.x,
+	       child->allocation.y,
+	       child->allocation.width,
+	       child->allocation.height,
+	       child_loc->x, child_loc->y,
+	       x, y);
+#endif /* DEBUG_TOOLTIP */
+
+      /* (x, y) relative to child's allocation. */
       if (x >= 0 && x < child->allocation.width
 	  && y >= 0 && y < child->allocation.height)
         {
@@ -428,12 +501,15 @@ child_location_foreach (GtkWidget *child,
 	    {
 	      struct ChildLocation tmp = { NULL, NULL, 0, 0 };
 
+	      /* Take (x, y) relative the child's allocation and
+	       * recurse.
+	       */
 	      tmp.x = x;
 	      tmp.y = y;
 	      tmp.container = child;
 
-	      gtk_container_foreach (GTK_CONTAINER (child),
-				     child_location_foreach, &tmp);
+	      gtk_container_forall (GTK_CONTAINER (child),
+				    child_location_foreach, &tmp);
 
 	      if (tmp.child)
 		child_loc->child = tmp.child;
@@ -446,6 +522,9 @@ child_location_foreach (GtkWidget *child,
     }
 }
 
+/* Translates coordinates from dest_widget->window relative (src_x, src_y),
+ * to allocation relative (dest_x, dest_y) of dest_widget.
+ */
 static void
 window_to_alloc (GtkWidget *dest_widget,
 		 gint       src_x,
@@ -459,6 +538,9 @@ window_to_alloc (GtkWidget *dest_widget,
       gint wx, wy;
       gdk_window_get_position (dest_widget->window, &wx, &wy);
 
+      /* Offset coordinates if widget->window is smaller than
+       * widget->allocation.
+       */
       src_x += wx - dest_widget->allocation.x;
       src_y += wy - dest_widget->allocation.y;
     }
@@ -474,6 +556,9 @@ window_to_alloc (GtkWidget *dest_widget,
     *dest_y = src_y;
 }
 
+/* Translates coordinates from window relative (x, y) to
+ * allocation relative (x, y) of the returned widget.
+ */
 static GtkWidget *
 find_widget_under_pointer (GdkWindow *window,
 			   gint      *x,
@@ -487,9 +572,20 @@ find_widget_under_pointer (GdkWindow *window,
   if (!event_widget)
     return NULL;
 
+#ifdef DEBUG_TOOLTIP
+  g_print ("event window %p (belonging to %p (%s))  (%d, %d)\n",
+	   window, event_widget, gtk_widget_get_name (event_widget),
+	   *x, *y);
+#endif
+
+  /* Coordinates are relative to event window */
   child_loc.x = *x;
   child_loc.y = *y;
 
+  /* We go down the window hierarchy to the widget->window,
+   * coordinates stay relative to the current window.
+   * We end up with window == widget->window, coordinates relative to that.
+   */
   while (window && window != event_widget->window)
     {
       gint px, py;
@@ -501,12 +597,6 @@ find_widget_under_pointer (GdkWindow *window,
       window = gdk_window_get_parent (window);
     }
 
-  if (GTK_WIDGET_NO_WINDOW (event_widget))
-    {
-      child_loc.x += event_widget->allocation.x;
-      child_loc.y += event_widget->allocation.y;
-    }
-
   /* Failing to find widget->window can happen for e.g. a detached handle box;
    * chaining ::query-tooltip up to its parent probably makes little sense,
    * and users better implement tooltips on handle_box->child.
@@ -515,24 +605,40 @@ find_widget_under_pointer (GdkWindow *window,
   if (!window)
     return NULL;
 
+  /* Convert the window relative coordinates to allocation
+   * relative coordinates.
+   */
+  window_to_alloc (event_widget,
+		   child_loc.x, child_loc.y,
+		   &child_loc.x, &child_loc.y);
+
   if (GTK_IS_CONTAINER (event_widget))
     {
-      window_to_alloc (event_widget,
-		       child_loc.x, child_loc.y,
-		       &child_loc.x, &child_loc.y);
+      GtkWidget *container = event_widget;
 
       child_loc.container = event_widget;
       child_loc.child = NULL;
 
-      gtk_container_foreach (GTK_CONTAINER (event_widget),
-			     child_location_foreach, &child_loc);
+      gtk_container_forall (GTK_CONTAINER (event_widget),
+			    child_location_foreach, &child_loc);
+
+      /* Here we have a widget, with coordinates relative to
+       * child_loc.container's allocation.
+       */
 
       if (child_loc.child)
 	event_widget = child_loc.child;
       else if (child_loc.container)
 	event_widget = child_loc.container;
+
+      /* Translate to event_widget's allocation */
+      gtk_widget_translate_coordinates (container, event_widget,
+					child_loc.x, child_loc.y,
+					&child_loc.x, &child_loc.y);
+
     }
 
+  /* We return (x, y) relative to the allocation of event_widget. */
   if (x)
     *x = child_loc.x;
   if (y)
@@ -541,6 +647,9 @@ find_widget_under_pointer (GdkWindow *window,
   return event_widget;
 }
 
+/* Ignores (x, y) on input, translates event coordinates to
+ * allocation relative (x, y) of the returned widget.
+ */
 static GtkWidget *
 find_topmost_widget_coords_from_event (GdkEvent *event,
 				       gint     *x,
@@ -554,8 +663,17 @@ find_topmost_widget_coords_from_event (GdkEvent *event,
   tx = dx;
   ty = dy;
 
+  /* Returns coordinates relative to tmp's allocation. */
   tmp = find_widget_under_pointer (event->any.window, &tx, &ty);
 
+  if (!tmp)
+    return NULL;
+
+  /* Make sure the pointer can actually be on the widget returned. */
+  if (tx < 0 || tx >= tmp->allocation.width ||
+      ty < 0 || ty >= tmp->allocation.height)
+    return NULL;
+
   if (x)
     *x = tx;
   if (y)
@@ -631,6 +749,84 @@ gtk_tooltip_run_requery (GtkWidget  **widget,
   return return_value;
 }
 
+static void
+gtk_tooltip_position (GtkTooltip *tooltip,
+		      GdkDisplay *display,
+		      GtkWidget  *new_tooltip_widget)
+{
+  gint x, y;
+  GdkScreen *screen;
+
+  tooltip->tooltip_widget = new_tooltip_widget;
+
+  /* Position the tooltip */
+  /* FIXME: should we swap this when RTL is enabled? */
+  if (tooltip->keyboard_mode_enabled)
+    {
+      gdk_window_get_origin (new_tooltip_widget->window, &x, &y);
+      if (GTK_WIDGET_NO_WINDOW (new_tooltip_widget))
+        {
+	  x += new_tooltip_widget->allocation.x;
+	  y += new_tooltip_widget->allocation.y;
+	}
+
+      /* For keyboard mode we position the tooltip below the widget,
+       * right of the center of the widget.
+       */
+      x += new_tooltip_widget->allocation.width / 2;
+      y += new_tooltip_widget->allocation.height + 4;
+    }
+  else
+    {
+      guint cursor_size;
+
+      x = tooltip->last_x;
+      y = tooltip->last_y;
+
+      /* For mouse mode, we position the tooltip right of the cursor,
+       * a little below the cursor's center.
+       */
+      cursor_size = gdk_display_get_default_cursor_size (display);
+      x += cursor_size / 2;
+      y += cursor_size / 2;
+    }
+
+  screen = gtk_widget_get_screen (new_tooltip_widget);
+
+  /* Show it */
+  if (tooltip->current_window)
+    {
+      gint monitor_num;
+      GdkRectangle monitor;
+      GtkRequisition requisition;
+
+      gtk_widget_size_request (GTK_WIDGET (tooltip->current_window),
+                               &requisition);
+
+      monitor_num = gdk_screen_get_monitor_at_point (screen, x, y);
+      gdk_screen_get_monitor_geometry (screen, monitor_num, &monitor);
+
+      if (x + requisition.width > monitor.x + monitor.width)
+        x -= x - (monitor.x + monitor.width) + requisition.width;
+      else if (x < monitor.x)
+        x = monitor.x;
+
+      if (y + requisition.height > monitor.y + monitor.height)
+        y -= y - (monitor.y + monitor.height) + requisition.height;
+  
+      if (!tooltip->keyboard_mode_enabled)
+        {
+          /* don't pop up under the pointer */
+          if (x <= tooltip->last_x && tooltip->last_x < x + requisition.width &&
+              y <= tooltip->last_y && tooltip->last_y < y + requisition.height)
+            y = tooltip->last_y - requisition.height - 2;
+        }
+  
+      gtk_window_move (GTK_WINDOW (tooltip->current_window), x, y);
+      gtk_widget_show (GTK_WIDGET (tooltip->current_window));
+    }
+}
+
 static void
 gtk_tooltip_show_tooltip (GdkDisplay *display)
 {
@@ -685,40 +881,9 @@ gtk_tooltip_show_tooltip (GdkDisplay *display)
 	tooltip->current_window = GTK_WINDOW (GTK_TOOLTIP (tooltip)->window);
     }
 
-  /* Position the tooltip */
-  /* FIXME: should we swap this when RTL is enabled? */
-  if (tooltip->keyboard_mode_enabled)
-    {
-      gdk_window_get_origin (tooltip_widget->window, &x, &y);
-      if (GTK_WIDGET_NO_WINDOW (tooltip_widget))
-        {
-	  x += tooltip_widget->allocation.x;
-	  y += tooltip_widget->allocation.y;
-	}
-
-      /* For keyboard mode we position the tooltip below the widget,
-       * right of the center of the widget.
-       */
-      x += tooltip_widget->allocation.width / 2;
-      y += tooltip_widget->allocation.height + 4;
-    }
-  else
-    {
-      guint cursor_size;
-
-      x = tooltip->last_x;
-      y = tooltip->last_y;
-
-      /* For mouse mode, we position the tooltip right of the cursor,
-       * a little below the cursor's center.
-       */
-      cursor_size = gdk_display_get_default_cursor_size (display);
-      x += cursor_size / 2;
-      y += cursor_size / 2;
-    }
-
   screen = gtk_widget_get_screen (tooltip_widget);
 
+  /* FIXME: should use tooltip->current_window iso tooltip->window */
   if (screen != gtk_widget_get_screen (tooltip->window))
     {
       g_signal_handlers_disconnect_by_func (display,
@@ -731,32 +896,7 @@ gtk_tooltip_show_tooltip (GdkDisplay *display)
 			G_CALLBACK (gtk_tooltip_display_closed), tooltip);
     }
 
-  tooltip->tooltip_widget = tooltip_widget;
-
-  /* Show it */
-  if (tooltip->current_window)
-    {
-      gint monitor_num;
-      GdkRectangle monitor;
-      GtkRequisition requisition;
-
-      gtk_widget_size_request (GTK_WIDGET (tooltip->current_window),
-                               &requisition);
-
-      monitor_num = gdk_screen_get_monitor_at_point (screen, x, y);
-      gdk_screen_get_monitor_geometry (screen, monitor_num, &monitor);
-
-      if (x + requisition.width > monitor.x + monitor.width)
-        x -= x - (monitor.x + monitor.width) + requisition.width;
-      else if (x < monitor.x)
-        x = monitor.x;
-
-      if (y + requisition.height > monitor.y + monitor.height)
-        y -= y - (monitor.y + monitor.height) + requisition.height;
-
-      gtk_window_move (GTK_WINDOW (tooltip->current_window), x, y);
-      gtk_widget_show (GTK_WIDGET (tooltip->current_window));
-    }
+  gtk_tooltip_position (tooltip, display, tooltip_widget);
 
   /* Now a tooltip is visible again on the display, make sure browse
    * mode is enabled.
@@ -997,10 +1137,22 @@ _gtk_tooltip_handle_event (GdkEvent *event)
 {
   gint x, y;
   gboolean return_value = FALSE;
+  gboolean touchscreen;
   GtkWidget *has_tooltip_widget = NULL;
+  GdkScreen *screen;
   GdkDisplay *display;
   GtkTooltip *current_tooltip;
+  GtkSettings *settings;
 
+  /* Disable tooltips in touchscreen mode */
+  screen = gdk_drawable_get_screen (event->any.window);
+  settings = gtk_settings_get_for_screen (screen);
+  g_object_get (settings, "gtk-touchscreen-mode", &touchscreen, NULL);
+
+  if (touchscreen)
+    return;
+
+  /* Returns coordinates relative to has_tooltip_widget's allocation. */
   has_tooltip_widget = find_topmost_widget_coords_from_event (event, &x, &y);
   display = gdk_drawable_get_display (event->any.window);
   current_tooltip = g_object_get_data (G_OBJECT (display),
@@ -1032,13 +1184,24 @@ _gtk_tooltip_handle_event (GdkEvent *event)
       return;
     }
 
+#ifdef DEBUG_TOOLTIP
+  if (has_tooltip_widget)
+    g_print ("%p (%s) at (%d, %d) %dx%d     pointer: (%d, %d)\n",
+	     has_tooltip_widget, gtk_widget_get_name (has_tooltip_widget),
+	     has_tooltip_widget->allocation.x,
+	     has_tooltip_widget->allocation.y,
+	     has_tooltip_widget->allocation.width,
+	     has_tooltip_widget->allocation.height,
+	     x, y);
+#endif /* DEBUG_TOOLTIP */
+
   /* Always poll for a next motion event */
   gdk_event_request_motions (&event->motion);
 
   /* Hide the tooltip when there's no new tooltip widget */
   if (!has_tooltip_widget)
     {
-      if (current_tooltip && GTK_TOOLTIP_VISIBLE (current_tooltip))
+      if (current_tooltip)
 	gtk_tooltip_hide_tooltip (current_tooltip);
 
       return;
@@ -1061,11 +1224,35 @@ _gtk_tooltip_handle_event (GdkEvent *event)
       case GDK_SCROLL:
 	if (current_tooltip)
 	  {
+	    gboolean tip_area_set;
+	    GdkRectangle tip_area;
+	    gboolean hide_tooltip;
+
+	    tip_area_set = current_tooltip->tip_area_set;
+	    tip_area = current_tooltip->tip_area;
+
 	    return_value = gtk_tooltip_run_requery (&has_tooltip_widget,
 						    current_tooltip,
 						    &x, &y);
 
-	    if (!return_value)
+	    /* Requested to be hidden? */
+	    hide_tooltip = !return_value;
+
+	    /* Leave notify should override the query function */
+	    hide_tooltip = (event->type == GDK_LEAVE_NOTIFY);
+
+	    /* Is the pointer above another widget now? */
+	    if (GTK_TOOLTIP_VISIBLE (current_tooltip))
+	      hide_tooltip |= has_tooltip_widget != current_tooltip->tooltip_widget;
+
+	    /* Did the pointer move out of the previous "context area"? */
+	    if (tip_area_set)
+	      hide_tooltip |= (x <= tip_area.x
+			       || x >= tip_area.x + tip_area.width
+			       || y <= tip_area.y
+			       || y >= tip_area.y + tip_area.height);
+
+	    if (hide_tooltip)
 	      gtk_tooltip_hide_tooltip (current_tooltip);
 	    else
 	      gtk_tooltip_start_delay (display);
